from pathlib import Path

from ._posix import home_dir as home_dir


def application_support_dir() -> Path:
    return home_dir() / "Library/Application Support"


def caches_dir() -> Path:
    return home_dir() / "Library/Caches"


def desktop_dir() -> Path:
    return home_dir() / "Desktop"


def documents_dir() -> Path:
    return home_dir() / "Documents"


def downloads_dir() -> Path:
    return home_dir() / "Downloads"


def movies_dir() -> Path:
    return home_dir() / "Movies"


def music_dir() -> Path:
    return home_dir() / "Music"


def pictures_dir() -> Path:
    return home_dir() / "Pictures"


def public_dir() -> Path:
    return home_dir() / "Public"



import os
from pathlib import Path

from .platforms import home_dir as home_dir

_USER_DIRS = None


def _base_dir(env_var, default):
    if (value := os.getenv(env_var)) is not None:
        if (path := Path(value)).is_absolute():
            return path
    return default


def _base_dirs(env_var, default):
    if (value := os.getenv(env_var)) is not None:
        paths = [
            path for dir in value.split(os.pathsep) if (path := Path(dir)).is_absolute()
        ]
        if paths:
            return paths
    return default


def _parse_user_dirs_file():
    user_dirs = {}
    user_dirs_path = config_dir() / "user-dirs.dirs"
    if user_dirs_path.exists():
        with user_dirs_path.open() as file:
            for line in file.readlines():
                if line.startswith("XDG_"):
                    key, value = line.split("=")
                    path = Path(
                        value.rstrip().strip('"').replace("$HOME", str(home_dir()), 1)
                    )
                    if path.is_absolute():
                        user_dirs[key] = path
    return user_dirs


def _user_dir(env_var):
    global _USER_DIRS
    _USER_DIRS = _parse_user_dirs_file() if _USER_DIRS is None else _USER_DIRS

    return _USER_DIRS.get(env_var, home_dir())


def bin_dir() -> Path:
    return _base_dir("XDG_BIN_HOME", home_dir() / ".local/bin")


def cache_dir() -> Path:
    return _base_dir("XDG_CACHE_HOME", home_dir() / ".cache")


def config_dir() -> Path:
    return _base_dir("XDG_CONFIG_HOME", home_dir() / ".config")


def data_dir() -> Path:
    return _base_dir("XDG_DATA_HOME", home_dir() / ".local/share")


def state_dir() -> Path:
    return _base_dir("XDG_STATE_HOME", home_dir() / ".local/state")


def runtime_dir() -> Path | None:
    return _base_dir("XDG_RUNTIME_DIR", None)


def data_dirs() -> list[Path]:
    return _base_dirs("XDG_DATA_DIRS", [Path("/usr/local/share"), Path("/usr/share")])


def config_dirs() -> list[Path]:
    return _base_dirs("XDG_CONFIG_DIRS", [Path("/etc/xdg")])


def desktop_dir() -> Path:
    return _user_dir("XDG_DESKTOP_DIR")


def documents_dir() -> Path:
    return _user_dir("XDG_DOCUMENTS_DIR")


def downloads_dir() -> Path:
    return _user_dir("XDG_DOWNLOAD_DIR")


def music_dir() -> Path:
    return _user_dir("XDG_MUSIC_DIR")


def pictures_dir() -> Path:
    return _user_dir("XDG_PICTURES_DIR")


def public_dir() -> Path:
    return _user_dir("XDG_PUBLICSHARE_DIR")


def templates_dir() -> Path:
    return _user_dir("XDG_TEMPLATES_DIR")


def videos_dir() -> Path:
    return _user_dir("XDG_VIDEOS_DIR")



import os
from pathlib import Path


def home_dir() -> Path:
    if (home := os.getenv("HOME")) is None:
        from pwd import getpwuid

        home = getpwuid(os.getuid()).pw_dir

    return Path(home)



from __future__ import annotations

import sys
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

if sys.platform == "win32":
    from .platforms.windows import (
        desktop_dir,
        documents_dir,
        downloads_dir,
        home_dir,
        music_dir,
        pictures_dir,
        public_dir,
        videos_dir,
    )
elif sys.platform == "darwin":
    from .platforms.macos import (
        desktop_dir,
        documents_dir,
        downloads_dir,
        home_dir,
        music_dir,
        pictures_dir,
        public_dir,
    )
    from .platforms.macos import movies_dir as videos_dir
else:
    from .xdg import (
        desktop_dir,
        documents_dir,
        downloads_dir,
        home_dir,
        music_dir,
        pictures_dir,
        public_dir,
        videos_dir,
    )

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator, Sequence

    PathOrSequencePath = TypeVar("PathOrSequencePath", bound=Path | Sequence[Path])

__all__ = [
    "iter_dirs",
    "app_name",
    "join_on",
    "dir_on",
    "cache_dir",
    "config_dir",
    "data_dir",
    "system_config_dirs",
    "system_data_dirs",
    "home_dir",
    "desktop_dir",
    "documents_dir",
    "downloads_dir",
    "music_dir",
    "pictures_dir",
    "public_dir",
    "videos_dir",
]


def iter_dirs(*args: Path | Sequence[Path]) -> Iterator[Path]:
    for arg in args:
        if isinstance(arg, Path):
            yield arg
        else:
            yield from arg


def _on(windows=None, macos=None, posix=None, others=None):
    if sys.platform == "win32":
        if windows is not None:
            return windows
    else:  # posix
        if sys.platform == "darwin":
            if macos is not None:
                return macos
        if posix is not None:
            return posix

    if others is not None:
        return others

    raise TypeError("at least 2 of `windows`, `posix`, and `others` must be given")


def app_name(
    app: str,
    org: str | None = None,
    tld: str | None = None,
    *,
    macos_asposix: bool = False,
) -> str:
    return _on(
        windows="\\".join(filter(None, [org, app])),
        macos=None
        if macos_asposix
        else ".".join(filter(None, [tld, org, app])).replace(" ", "-"),
        posix=app.replace(" ", "").lower(),
    )


def join_on(
    path: str | Path,
    *,
    windows: str | Path | None = None,
    macos: str | Path | None = None,
    posix: str | Path | None = None,
    others: str | Path | None = None,
) -> Path:
    return Path(path, _on(windows, macos, posix, others))


def dir_on(
    *,
    windows: Callable[..., PathOrSequencePath] | None = None,
    macos: Callable[..., PathOrSequencePath] | None = None,
    posix: Callable[..., PathOrSequencePath] | None = None,
    others: Callable[..., PathOrSequencePath] | None = None,
) -> PathOrSequencePath:
    return _on(windows, macos, posix, others)()


def _windows_data_dir(system, roaming=True):
    from .platforms.windows import (
        GUID,
        app_data_local_dir,
        app_data_roaming_dir,
        known_folder,
    )

    return (
        known_folder(GUID.from_int(0x62AB5D82_FDC1_4DC3_A9DD_070D1D495D97))
        if system
        else (app_data_roaming_dir() if roaming else app_data_local_dir())
    )


def _macos_cache_dir(system):
    from .platforms.macos import caches_dir

    return Path("/Library/Caches") if system else caches_dir()


def _macos_data_dir(system):
    from .platforms.macos import application_support_dir

    return Path("/Library/Application Support") if system else application_support_dir()


def _posix_cache_dir(system):
    from .xdg import cache_dir

    return Path("/var/cache") if system else cache_dir()


def _posix_config_dir(system):
    from .xdg import config_dir

    return Path("/etc") if system else config_dir()


def _posix_data_dir(system):
    from .xdg import data_dir

    return Path("/usr/share") if system else data_dir()


def cache_dir(*, system: bool = False, macos_asposix: bool = False) -> Path:
    return dir_on(
        windows=lambda: _windows_data_dir(system, roaming=False),
        macos=None if macos_asposix else lambda: _macos_cache_dir(system),
        posix=lambda: _posix_cache_dir(system),
    )


def config_dir(
    *, system: bool = False, windows_roaming: bool = True, macos_asposix: bool = False
) -> Path:
    return dir_on(
        windows=lambda: _windows_data_dir(system, windows_roaming),
        macos=None if macos_asposix else lambda: _macos_data_dir(system),
        posix=lambda: _posix_config_dir(system),
    )


def data_dir(
    *, system: bool = False, windows_roaming: bool = True, macos_asposix: bool = False
) -> Path:
    return dir_on(
        windows=lambda: _windows_data_dir(system, windows_roaming),
        macos=None if macos_asposix else lambda: _macos_data_dir(system),
        posix=lambda: _posix_data_dir(system),
    )


def system_config_dirs(*, macos_asposix: bool = False) -> list[Path]:
    return dir_on(
        windows=lambda: [_windows_data_dir(system=True)],
        macos=None if macos_asposix else lambda: [_macos_data_dir(system=True)],
        posix=lambda: [Path("/usr/local/etc"), Path("/etc")],
    )


def system_data_dirs(*, macos_asposix: bool = False) -> list[Path]:
    return dir_on(
        windows=lambda: [_windows_data_dir(system=True)],
        macos=None if macos_asposix else lambda: [_macos_data_dir(system=True)],
        posix=lambda: [Path("/usr/local/share"), Path("/usr/share")],
    )



import ctypes
import sys
from ctypes import wintypes
from pathlib import Path


class GUID(ctypes.Structure):
    _fields_ = [
        ("Data1", wintypes.DWORD),
        ("Data2", wintypes.WORD),
        ("Data3", wintypes.WORD),
        ("Data4", wintypes.BYTE * 8),
    ]

    @classmethod
    def from_str(cls, guid: str):
        guid = guid.strip("{}").replace("-", "")

        return cls.from_int(int(guid, 16))

    @classmethod
    def from_int(cls, guid: int):
        data1 = (guid >> 96) & 0xFFFFFFFF
        data2 = (guid >> 80) & 0xFFFF
        data3 = (guid >> 64) & 0xFFFF
        data4 = tuple((guid & 0xFFFFFFFFFFFFFFFF).to_bytes(8, byteorder="big"))

        return cls(data1, data2, data3, data4)


if sys.platform == "win32":
    _CoTaskMemFree = ctypes.windll.ole32.CoTaskMemFree
    _SHGetKnownFolderPath = ctypes.windll.shell32.SHGetKnownFolderPath

    _CoTaskMemFree.restype = None
    _CoTaskMemFree.argtypes = [wintypes.LPVOID]

    _SHGetKnownFolderPath.restype = wintypes.LONG
    _SHGetKnownFolderPath.argtypes = [
        ctypes.POINTER(GUID),
        wintypes.DWORD,
        wintypes.HANDLE,
        ctypes.POINTER(wintypes.LPWSTR),
    ]


def known_folder(guid: GUID) -> Path:
    ptr = wintypes.LPWSTR()
    hresult = _SHGetKnownFolderPath(ctypes.byref(guid), 0, None, ctypes.byref(ptr))

    try:
        if hresult == 0:
            return Path(ptr.value)  # type: ignore
        else:
            raise ctypes.WinError(hresult)
    finally:
        _CoTaskMemFree(ptr)


def home_dir() -> Path:
    return known_folder(GUID.from_int(0x5E6C858F_0E22_4760_9AFE_EA3317B67173))


def app_data_local_dir() -> Path:
    return known_folder(GUID.from_int(0xF1B32785_6FBA_4FCF_9D55_7B8E7F157091))


def app_data_roaming_dir() -> Path:
    return known_folder(GUID.from_int(0x3EB685DB_65F9_4CF6_A03A_E3EF65729F3D))


def desktop_dir() -> Path:
    return known_folder(GUID.from_int(0xB4BFCC3A_DB2C_424C_B029_7FE99A87C641))


def documents_dir() -> Path:
    return known_folder(GUID.from_int(0xFDD39AD0_238F_46AF_ADB4_6C85480369C7))


def downloads_dir() -> Path:
    return known_folder(GUID.from_int(0x374DE290_123F_4565_9164_39C4925E467B))


def music_dir() -> Path:
    return known_folder(GUID.from_int(0x4BD8D571_6D19_48D3_BE97_422220080E43))


def pictures_dir() -> Path:
    return known_folder(GUID.from_int(0x33E28130_4E1E_4676_835A_98395C3BC3BB))


def public_dir() -> Path:
    return known_folder(GUID.from_int(0xDFDF76A2_C82A_4D63_906A_5644AC457385))


def videos_dir() -> Path:
    return known_folder(GUID.from_int(0x18989B1D_99B5_455B_841C_AB7C74E4DDFC))
